import json

from replay.models import Action

def escape(text):
    "Escape text for string.Template substitution."
    return text.replace('$', '$$')

class RecorderMiddleware(object):
    def process_response(self, request, response):
        "Create Action object based on request and response."
        method = request.method
        data = json.dumps(getattr(request, method))
        status_code = response.status_code
        redirect = 300 <= status_code < 400
        content = response.content if not redirect else response.url

        try:
            content.decode('utf-8')
        except UnicodeDecodeError:
            content = ''

        Action.objects.create(
            method=method,
            path=escape(request.path),
            data=escape(data),
            status_code=str(status_code),
            content=content,
        )

        return response



import re
import time

from django import forms
from django.contrib import admin
from django.core.urlresolvers import reverse
from django.shortcuts import redirect
from django.utils.html import format_html, format_html_join
from django.utils.safestring import mark_safe

from replay.models import Action, Validator, Scenario, Step

class ActionAdminForm(forms.ModelForm):
    class Meta:
        model = Action
        fields = []  # Required but ignored.
        widgets = {
            'name': forms.TextInput(attrs={'size': 80}),
            'path': forms.TextInput(attrs={'size': 80}),
        }

class ValidatorInline(admin.StackedInline):
    model = Validator
    extra = 10
    ordering = (
        'order',
        'id',
    )
    fields = (
        'order',
        'pattern',
    )

def create_scenario(modeladmin, request, queryset):
    scenario = Scenario(name='Scenario %s' % time.ctime())
    scenario.save()
    for action in queryset:
        step = Step(scenario=scenario, action=action)
        step.save()
    return redirect('admin:replay_scenario_change', scenario.id)

create_scenario.short_description = 'Create scenario from actions'

def create_validators(modeladmin, request, queryset):
    title_pattern = '<title>(.*?)</title>'

    for action in queryset:
        pattern = None

        if action.status_code.startswith('3'):
            pattern = action.content
        else:
            match = re.search(title_pattern, action.content)
            if match:
                title = match.group(1)
                pattern = re.escape(title)
                if action.name == '':
                    action.name = title
                    action.save()

        if pattern:
            Validator.objects.create(
                action=action,
                pattern=pattern,
            )

create_validators.short_description = 'Create validators automatically'

class ActionAdmin(admin.ModelAdmin):
    actions = [
        create_scenario,
        create_validators,
    ]
    fields = (
        'name',
        ('method', 'path'),
        'data',
        'status_code',
        'content',
    )
    form = ActionAdminForm
    inlines = (
        ValidatorInline,
    )
    list_display = (
        'id',
        'name',
        'method',
        'path',
        'status_code',
    )
    ordering = ('name', 'id')
    save_on_top = True
    search_fields = (
        'name',
        'method',
        'path',
        'status_code',
    )

admin.site.register(Action, ActionAdmin)

class ValidatorAdmin(admin.ModelAdmin):
    fields = (
        'action',
        'order',
        'pattern',
    )
    list_display = (
        'id',
        'action',
        'order',
        '__str__',
    )
    ordering = (
        'action',
        'order',
        'id',
    )
    save_on_top = True
    search_fields = (
        'action__name',
        'action__path',
        'pattern',
    )

admin.site.register(Validator, ValidatorAdmin)

class StepInline(admin.StackedInline):
    model = Step
    extra = 10
    ordering = (
        'order',
        'id',
    )
    fields = (
        'order',
        'action',
        'action_url',
        'action_method',
        'action_path',
        'action_data',
        'action_status_code',
        'action_validators',
    )

    def action_url(self, step):
        return format_html(
            '<a href="{}">{}</a>',
            reverse('admin:replay_action_change', args=(step.action.id,)),
            str(step.action),
        )

    def action_method(self, step):
        return step.action.method

    def action_path(self, step):
        return step.action.path

    def action_data(self, step):
        return step.action.data

    def action_status_code(self, step):
        return step.action.status_code

    def action_validators(self, step):
        validators = Validator.objects.filter(action=step.action)
        validators = validators.order_by('order', 'id')
        iterable = ((validator.pattern,) for validator in validators)
        return format_html_join(mark_safe('<br><br>'), '&bull; {}', iterable)

    readonly_fields = (
        'action_url',
        'action_method',
        'action_path',
        'action_data',
        'action_status_code',
        'action_validators',
    )

class ScenarioAdminForm(forms.ModelForm):
    class Meta:
        model = Scenario
        fields = []  # Required but ignored.
        widgets = {
            'name': forms.TextInput(attrs={'size': 80}),
        }

class ScenarioAdmin(admin.ModelAdmin):
    fields = (
        'name',
        'priority',
    )
    form = ScenarioAdminForm
    inlines = (
        StepInline,
    )
    list_display = (
        'name',
        'priority',
    )
    save_on_top = True
    search_fields = (
        'name',
    )

admin.site.register(Scenario, ScenarioAdmin)

class StepAdmin(admin.ModelAdmin):
    fields = (
        'scenario',
        'action',
        'order',
    )
    list_display = (
        'id',
        'scenario',
        'action',
    )
    ordering = (
        'scenario__name',
        'order',
        'id',
    )
    save_on_top = True
    search_fields = (
        'scenario__name',
        'action__name',
        'action__path',
    )

admin.site.register(Step, StepAdmin)



"""Test Scenario Management Command

"""

import json
import re
import string
import urlparse

from django.core.management.base import BaseCommand, CommandError
from django.test import Client

from replay.models import Validator, Scenario, Step

def expand(text, mapping):
    template = string.Template(text)
    return template.substitute(mapping)


class Command(BaseCommand):
    can_import_settings = True

    def handle(self, *args, **options):
        try:
            scenarios = []
            for name in args:
                scenario = Scenario.objects.get(name=name)
                scenarios.append(scenario)
        except Scenario.DoesNotExist:
            message = 'scenario with name %r does not exist' % name
            raise CommandError(message)
        else:
            if not scenarios:
                scenarios = Scenario.objects.all().order_by('priority', 'id')

        self.client = Client()
        state = {}

        for scenario in scenarios:
            self.run(scenario, state)


    def run(self, scenario, state):
        steps = Step.objects.filter(scenario=scenario).order_by('order', 'id')
        errors = 0

        for step in steps:
            action = step.action
            status_code, content = self.request(action, state)
            
            if status_code != action.status_code:
                message = 'FAIL %r status code: %s expected: %s'
                data = action, status_code, action.status_code
                self.stdout.write(message % data)
                self.stdout.write('FAIL STATE')
                for key, value in state.items():
                    self.stdout.write('FAIL %s = %s' % (key, value))
                
                errors += 1
                break

            validators = Validator.objects.filter(action=action)
            validators = validators.order_by('order', 'id')
            
            for validator in validators:
                pattern = expand(validator.pattern, state)
                match = re.search(pattern, content)

                if match:
                    message = 'PASS %r %r %r' % (scenario, action, validator)
                    self.stdout.write(message)
                    state.update(match.groupdict())
                else:
                    message = 'FAIL %r %r %r' % (scenario, action, validator)
                    self.stdout.write(message)
                    errors += 1

        if errors:
            raise CommandError('%s errors' % errors)


    def request(self, action, state):
        func = getattr(self.client, action.method.lower())
        data = json.loads(expand(action.data, state))
        path = expand(action.path, state)
        response = func(path, data)
        status_code = response.status_code

        if 300 <= status_code < 400:
            return str(status_code), response.url
        else:
            return str(status_code), response.content



"""Record Management Command

TODO

* Accept optional [scenario name] argument to record scenario.
* When recording scenario, set "order" field on each step.

"""

from django.core.management import call_command
from django.core.management.base import BaseCommand

class Command(BaseCommand):
    can_import_settings = True

    def handle(self, *args, **options):
        from django.conf import settings
        settings.MIDDLEWARE_CLASSES = (
            ('replay.middleware.RecorderMiddleware',)
            + settings.MIDDLEWARE_CLASSES
        )
        call_command('runserver')



from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import sys
import replay


class Tox(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True
    def run_tests(self):
        import tox
        errno = tox.cmdline(self.test_args)
        sys.exit(errno)


with open('README.rst') as reader:
    readme = reader.read()

setup(
    name='django-replay',
    version=replay.__version__,
    description='Record and replay web requests.',
    long_description=readme,
    author='Grant Jenks',
    author_email='contact@grantjenks.com',
    url='http://www.grantjenks.com/docs/django-replay/',
    py_modules=['replay'],
    package_data={'': ['LICENSE', 'README.rst']},
    tests_require=['tox'],
    cmdclass={'test': Tox},
    license='Apache 2.0',
    install_requires=[],
    classifiers=(
        # 'Development Status :: 4 - Beta',
        # 'Intended Audience :: Developers',
        # 'License :: OSI Approved :: Apache Software License',
        # 'Natural Language :: English',
        # 'Programming Language :: Python',
        # 'Programming Language :: Python :: 2',
        # 'Programming Language :: Python :: 2.7',
        # 'Programming Language :: Python :: Implementation :: CPython',
    ),
)



"""Django Replay Models

* Remember to use dumpdata and loaddata!
* Do not make actions into templates/generics without great care! Some data
  like files may be deleted because they are stored outside the database.

TODO

* Admin function to run validator on response.

"""

from django.db import models

HTTP_METHODS = (
    ('GET', 'GET'),
    ('POST', 'POST'),
)

class Action(models.Model):
    name = models.TextField(blank=True, default='')
    method = models.CharField(max_length=8, choices=HTTP_METHODS)
    path = models.TextField()
    data = models.TextField()
    status_code = models.CharField(max_length=3)
    content = models.TextField(blank=True)

    def __str__(self):
        fields = (self.id, self.method, self.path)
        return self.name or '<Action: %s %s %s>' % fields

class Validator(models.Model):
    action = models.ForeignKey(Action)
    order = models.FloatField(default=1.0)
    pattern = models.TextField()

    def __str__(self):
        return self.pattern[:78]

class Scenario(models.Model):
    name = models.TextField(unique=True)
    priority = models.FloatField(default=1.0)

    def __str__(self):
        return self.name

class Step(models.Model):
    scenario = models.ForeignKey(Scenario)
    order = models.FloatField(default=1.0)
    action = models.ForeignKey(Action)

    def __str__(self):
        return '<Step: %s>' % self.id
